@@ -14,8 +14,9 @@ use ssbh_lib::{
 use ssbh_lib::{Half, Matrix3x3, SsbhArray, Vector3};
 use std::collections::{HashMap, HashSet};
 use std::convert::TryFrom;
-use std::io::Seek;
+use std::io::{Read, Seek};
 use std::ops::{Add, Div, Sub};
+use std::path::Path;
 use std::{error::Error, io::Write};
 
 use crate::{read_data, read_vector_data, write_f16, write_f32, write_u8, write_vector_data};
@@ -389,6 +390,60 @@ pub struct BoneInfluence {
     pub vertex_weights: Vec<VertexWeight>,
 }
 
+/// The data associated with a [Mesh] file.
+/// Supported versions are 1.8 and 1.10.
+#[derive(Debug, Clone)]
+pub struct MeshData {
+    pub major_version: u16,
+    pub minor_version: u16,
+    pub objects: Vec<MeshObjectData>,
+}
+
+// TODO: Restrict the error type?
+// TODO: Make this a trait?
+impl MeshData {
+    /// Tries to read and convert the MESH from `path`.
+    /// The entire file is buffered for performance.
+    pub fn from_file<P: AsRef<Path>>(path: P) -> Result<Self, Box<dyn std::error::Error>> {
+        let mesh = Mesh::from_file(path)?;
+        Ok(MeshData {
+            major_version: mesh.major_version,
+            minor_version: mesh.minor_version,
+            objects: read_mesh_objects(&mesh)?,
+        })
+    }
+
+    /// Tries to read and convert the MESH from `reader`.
+    /// For best performance when opening from a file, use `from_file` instead.
+    pub fn read<R: Read + Seek>(reader: &mut R) -> Result<Self, Box<dyn std::error::Error>> {
+        let mesh = Mesh::read(reader)?;
+        Ok(MeshData {
+            major_version: mesh.major_version,
+            minor_version: mesh.minor_version,
+            objects: read_mesh_objects(&mesh)?,
+        })
+    }
+
+    /// Converts the data to MESH and writes to the given `writer`.
+    /// For best performance when writing to a file, use `write_to_file` instead.
+    pub fn write<W: std::io::Write + Seek>(
+        &self,
+        writer: &mut W,
+    ) -> Result<(), Box<dyn std::error::Error>> {
+        let mesh = create_mesh(&self)?;
+        mesh.write(writer)?;
+        Ok(())
+    }
+
+    /// Converts the data to MESH and writes to the given `path`.
+    /// The entire file is buffered for performance.
+    pub fn write_to_file<P: AsRef<Path>>(&self, path: P) -> Result<(), Box<dyn std::error::Error>> {
+        let mesh = create_mesh(&self)?;
+        mesh.write_to_file(path)?;
+        Ok(())
+    }
+}
+
 #[derive(Debug, Clone)]
 pub struct MeshObjectData {
     pub name: String,
@@ -413,7 +468,7 @@ pub struct AttributeData {
     pub data: VectorData,
 }
 
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, PartialEq)]
 pub enum VectorData {
     Vector2(Vec<[f32; 2]>),
     Vector3(Vec<[f32; 3]>),
@@ -422,11 +477,13 @@ pub enum VectorData {
 
 impl VectorData {
     /// The number of vectors.
-    /// ```rust
-    /// # use ssbh_data::mesh_data::VectorData;
-    /// let data = VectorData::Vector2(vec![[0f32, 1f32], [0f32, 1f32], [0f32, 1f32]]);
-    /// assert_eq!(3, data.len());
-    /// ```
+    /**
+    ```rust
+    # use ssbh_data::mesh_data::VectorData;
+    let data = VectorData::Vector2(vec![[0f32, 1f32], [0f32, 1f32], [0f32, 1f32]]);
+    assert_eq!(3, data.len());
+    ```
+    */
     pub fn len(&self) -> usize {
         match self {
             VectorData::Vector2(v) => v.len(),
@@ -435,6 +492,11 @@ impl VectorData {
         }
     }
 
+    /// Returns `true` if there are no elements.
+    pub fn is_empty(&self) -> bool {
+        self.len() == 0
+    }
+
     fn write<W: Write + Seek, F: Fn(&mut W, &[f32]) -> std::io::Result<()>>(
         &self,
         writer: &mut W,
@@ -448,8 +510,44 @@ impl VectorData {
             VectorData::Vector4(v) => write_vector_data(writer, v, offset, stride, write_t),
         }
     }
+
+    fn to_vec3a(&self) -> Vec<geometry_tools::glam::Vec3A> {
+        match self {
+            VectorData::Vector2(data) => data
+                .iter()
+                .map(|[x, y]| geometry_tools::glam::Vec3A::new(*x, *y, 0f32))
+                .collect(),
+            VectorData::Vector3(data) => data
+                .iter()
+                .map(|[x, y, z]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
+                .collect(),
+            VectorData::Vector4(data) => data
+                .iter()
+                .map(|[x, y, z, _]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
+                .collect(),
+        }
+    }
+
+    fn to_vec4_with_w(&self, w: f32) -> Vec<geometry_tools::glam::Vec4> {
+        // Allow conversion to homogeneous coordinates by specifying the w component.
+        match self {
+            VectorData::Vector2(data) => data
+                .iter()
+                .map(|[x, y]| geometry_tools::glam::Vec4::new(*x, *y, 0f32, w))
+                .collect(),
+            VectorData::Vector3(data) => data
+                .iter()
+                .map(|[x, y, z]| geometry_tools::glam::Vec4::new(*x, *y, *z, w))
+                .collect(),
+            VectorData::Vector4(data) => data
+                .iter()
+                .map(|[x, y, z, _]| geometry_tools::glam::Vec4::new(*x, *y, *z, w))
+                .collect(),
+        }
+    }
 }
 
+// TODO: Create a MeshData struct and do MeshData::try_from(mesh)?
 pub fn read_mesh_objects(mesh: &Mesh) -> Result<Vec<MeshObjectData>, Box<dyn Error>> {
     let mut mesh_objects = Vec::new();
 
@@ -497,35 +595,33 @@ enum MeshVersion {
     Version108,
 }
 
-pub fn create_mesh(
-    major_version: u16,
-    minor_version: u16,
-    object_data: &[MeshObjectData],
-) -> Result<Mesh, MeshError> {
-    let version = match (major_version, minor_version) {
+// TODO: This can be a method on MeshData.
+pub fn create_mesh(data: &MeshData) -> Result<Mesh, MeshError> {
+    let version = match (data.major_version, data.minor_version) {
         (1, 10) => Ok(MeshVersion::Version110),
         (1, 8) => Ok(MeshVersion::Version108),
         _ => Err(MeshError::UnsupportedMeshVersion {
-            major_version,
-            minor_version,
+            major_version: data.major_version,
+            minor_version: data.minor_version,
         }),
     }?;
 
-    let mesh_vertex_data = create_mesh_objects(version, object_data)?;
+    let mesh_vertex_data = create_mesh_objects(version, &data.objects)?;
 
     // TODO: It might be more efficient to reuse the data for mesh object bounding or reuse the generated points.
-    let all_positions: Vec<geometry_tools::glam::Vec3A> = object_data
+    let all_positions: Vec<geometry_tools::glam::Vec3A> = data
+        .objects
         .iter()
         .map(|o| match o.positions.first() {
-            Some(attribute) => vector_data_to_glam(&attribute.data),
+            Some(attribute) => attribute.data.to_vec3a(),
             None => Vec::new(),
         })
         .flatten()
         .collect();
 
     let mesh = Mesh {
-        major_version,
-        minor_version,
+        major_version: data.major_version,
+        minor_version: data.minor_version,
         model_name: "".into(),
         bounding_info: calculate_bounding_info(&all_positions),
         unk1: 0,
@@ -542,11 +638,11 @@ pub fn create_mesh(
         vertex_buffers: mesh_vertex_data
             .vertex_buffers
             .into_iter()
-            .map(|b| SsbhByteBuffer::new(b))
+            .map(SsbhByteBuffer::new)
             .collect::<Vec<SsbhByteBuffer>>()
             .into(),
         index_buffer: mesh_vertex_data.index_buffer.into(),
-        rigging_buffers: create_rigging_buffers(version, object_data)?.into(),
+        rigging_buffers: create_rigging_buffers(version, &data.objects)?.into(),
     };
 
     Ok(mesh)
@@ -560,7 +656,7 @@ fn calculate_max_influences(influences: &[BoneInfluence]) -> usize {
             // Assume influences are uniquely identified by their bone name.
             let entry = influences_by_vertex
                 .entry(weight.vertex_index)
-                .or_insert(HashSet::new());
+                .or_insert_with(HashSet::new);
             entry.insert(&influence.bone_name);
         }
     }
@@ -941,7 +1037,7 @@ fn create_mesh_objects(
         // Assume generated bounding data isn't critical if there are no points.
         // Use an empty list if there are no position attributes.
         let positions = match data.positions.first() {
-            Some(attribute) => vector_data_to_glam(&attribute.data),
+            Some(attribute) => attribute.data.to_vec3a(),
             None => Vec::new(),
         };
 
@@ -1193,21 +1289,108 @@ fn convert_indices(indices: &[u32]) -> VertexIndices {
     }
 }
 
-fn vector_data_to_glam(points: &VectorData) -> Vec<geometry_tools::glam::Vec3A> {
-    match points {
-        VectorData::Vector2(data) => data
-            .iter()
-            .map(|[x, y]| geometry_tools::glam::Vec3A::new(*x, *y, 0f32))
-            .collect(),
-        VectorData::Vector3(data) => data
-            .iter()
-            .map(|[x, y, z]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
-            .collect(),
-        VectorData::Vector4(data) => data
-            .iter()
-            .map(|[x, y, z, _]| geometry_tools::glam::Vec3A::new(*x, *y, *z))
-            .collect(),
+fn transform_inner(data: &VectorData, transform: &[[f32; 4]; 4], w: f32) -> VectorData {
+    let mut points = data.to_vec4_with_w(w);
+
+    // Transform is assumed to be row-major.
+    // Skip tranposing when converting to ensure the correct result inside the loop.
+    let matrix = glam::Mat4::from_cols_array_2d(transform);
+    for point in points.iter_mut() {
+        *point = matrix.mul_vec4(*point);
     }
+
+    // Preserve the original component count.
+    match data {
+        VectorData::Vector2(_) => VectorData::Vector2(points.iter().map(|p| [p.x, p.y]).collect()),
+        VectorData::Vector3(_) => {
+            VectorData::Vector3(points.iter().map(|p| [p.x, p.y, p.z]).collect())
+        }
+        // Preserve the original w component.
+        // For example, tangents often store a sign component in the w component.
+        VectorData::Vector4(original) => VectorData::Vector4(
+            original
+                .iter()
+                .zip(points)
+                .map(|(old, new)| [new.x, new.y, new.z, old[3]])
+                .collect(),
+        ),
+    }
+}
+
+/// Transform the elements in `data` with `transform`.
+/// Transform is assumed to be in row-major order.
+/// The elements are treated as points in homogeneous coordinates by temporarily setting the 4th component to `1.0f32`.
+/// The returned result has the same component count as `data`.
+/// For [VectorData::Vector4], the 4th component is preserved for the returned result.
+/**
+```rust
+# use ssbh_data::mesh_data::{VectorData, AttributeData, MeshObjectData, transform_points};
+# let mesh_object_data = MeshObjectData {
+#     name: "abc".into(),
+#     sub_index: 0,
+#     parent_bone_name: "".into(),
+#     vertex_indices: Vec::new(),
+#     positions: vec![AttributeData {
+#         name: "Position0".into(),
+#         data: VectorData::Vector3(Vec::new())
+#     }],
+#     normals: Vec::new(),
+#     binormals: Vec::new(),
+#     tangents: Vec::new(),
+#     texture_coordinates: Vec::new(),
+#     color_sets: Vec::new(),
+#     bone_influences: Vec::new(),
+# };
+// A scaling matrix for x, y, and z.
+let transform = [
+    [1f32, 0f32, 0f32, 0f32],
+    [0f32, 2f32, 0f32, 0f32],
+    [0f32, 0f32, 3f32, 0f32],
+    [0f32, 0f32, 0f32, 1f32],
+];
+let transformed_positions = transform_points(&mesh_object_data.positions[0].data, &transform);
+```
+*/
+pub fn transform_points(data: &VectorData, transform: &[[f32; 4]; 4]) -> VectorData {
+    transform_inner(data, transform, 1.0)
+}
+
+/// Transform the elements in `data` with `transform`.
+/// Transform is assumed to be in row-major order.
+/// The elements are treated as vectors in homogeneous coordinates by temporarily setting the 4th component to `0.0f32`.
+/// The returned result has the same component count as `data`.
+/// For [VectorData::Vector4], the 4th component is preserved for the returned result.
+/**
+```rust
+# use ssbh_data::mesh_data::{VectorData, AttributeData, MeshObjectData, transform_vectors};
+# let mesh_object_data = MeshObjectData {
+#     name: "abc".into(),
+#     sub_index: 0,
+#     parent_bone_name: "".into(),
+#     vertex_indices: Vec::new(),
+#     positions: Vec::new(),
+#     normals: vec![AttributeData {
+#         name: "Normal0".into(),
+#         data: VectorData::Vector3(Vec::new())
+#     }],
+#     binormals: Vec::new(),
+#     tangents: Vec::new(),
+#     texture_coordinates: Vec::new(),
+#     color_sets: Vec::new(),
+#     bone_influences: Vec::new(),
+# };
+// A scaling matrix for x, y, and z.
+let transform = [
+    [1f32, 0f32, 0f32, 0f32],
+    [0f32, 2f32, 0f32, 0f32],
+    [0f32, 0f32, 3f32, 0f32],
+    [0f32, 0f32, 0f32, 1f32],
+];
+let transformed_normals = transform_vectors(&mesh_object_data.normals[0].data, &transform);
+```
+*/
+pub fn transform_vectors(data: &VectorData, transform: &[[f32; 4]; 4]) -> VectorData {
+    transform_inner(data, transform, 0.0)
 }
 
 fn calculate_bounding_info(
@@ -1667,10 +1850,7 @@ mod tests {
                 assert_eq!(0, a.sub_index);
                 assert_eq!(AttributeDataType::HalfFloat2, a.data_type);
                 assert_eq!("firstUv", a.name.to_str().unwrap());
-                assert_eq!(
-                    "firstUv",
-                    a.attribute_names.elements[0].to_str().unwrap()
-                );
+                assert_eq!("firstUv", a.attribute_names.elements[0].to_str().unwrap());
 
                 let a = attributes.next().unwrap();
                 assert_eq!(AttributeUsageV10::TextureCoordinate, a.usage);
@@ -1679,10 +1859,7 @@ mod tests {
                 assert_eq!(1, a.sub_index);
                 assert_eq!(AttributeDataType::HalfFloat2, a.data_type);
                 assert_eq!("secondUv", a.name.to_str().unwrap());
-                assert_eq!(
-                    "secondUv",
-                    a.attribute_names.elements[0].to_str().unwrap()
-                );
+                assert_eq!("secondUv", a.attribute_names.elements[0].to_str().unwrap());
 
                 let a = attributes.next().unwrap();
                 assert_eq!(AttributeUsageV10::ColorSet, a.usage);
@@ -1691,10 +1868,7 @@ mod tests {
                 assert_eq!(0, a.sub_index);
                 assert_eq!(AttributeDataType::Byte4, a.data_type);
                 assert_eq!("color1", a.name.to_str().unwrap());
-                assert_eq!(
-                    "color1",
-                    a.attribute_names.elements[0].to_str().unwrap()
-                );
+                assert_eq!("color1", a.attribute_names.elements[0].to_str().unwrap());
 
                 let a = attributes.next().unwrap();
                 assert_eq!(AttributeUsageV10::ColorSet, a.usage);
@@ -1703,10 +1877,7 @@ mod tests {
                 assert_eq!(1, a.sub_index);
                 assert_eq!(AttributeDataType::Byte4, a.data_type);
                 assert_eq!("color2", a.name.to_str().unwrap());
-                assert_eq!(
-                    "color2",
-                    a.attribute_names.elements[0].to_str().unwrap()
-                );
+                assert_eq!("color2", a.attribute_names.elements[0].to_str().unwrap());
             }
             _ => panic!("invalid version"),
         };
@@ -2023,7 +2194,12 @@ mod tests {
 
     #[test]
     fn create_empty_mesh_1_10() {
-        let mesh = create_mesh(1, 10, &[]).unwrap();
+        let mesh = create_mesh(&MeshData {
+            major_version: 1,
+            minor_version: 10,
+            objects: Vec::new(),
+        })
+        .unwrap();
         assert_eq!(1, mesh.major_version);
         assert_eq!(10, mesh.minor_version);
         assert!(mesh.objects.elements.is_empty());
@@ -2033,7 +2209,13 @@ mod tests {
 
     #[test]
     fn create_empty_mesh_1_8() {
-        let mesh = create_mesh(1, 8, &[]).unwrap();
+        let mesh = create_mesh(&MeshData {
+            major_version: 1,
+            minor_version: 8,
+            objects: Vec::new(),
+        })
+        .unwrap();
+
         assert_eq!(1, mesh.major_version);
         assert_eq!(8, mesh.minor_version);
         assert!(mesh.objects.elements.is_empty());
@@ -2044,6 +2226,70 @@ mod tests {
     #[test]
     #[should_panic(expected = "Creating a version 2.301 mesh is not supported.")]
     fn create_empty_mesh_invalid_version() {
-        create_mesh(2, 301, &[]).unwrap();
+        create_mesh(&MeshData {
+            major_version: 2,
+            minor_version: 301,
+            objects: Vec::new(),
+        })
+        .unwrap();
+    }
+
+    #[test]
+    fn transform_points_vec2() {
+        let data = VectorData::Vector2(vec![[0f32, 1f32], [2f32, 3f32]]);
+        let transform = [
+            [2f32, 0f32, 0f32, 0f32],
+            [0f32, 3f32, 0f32, 0f32],
+            [0f32, 0f32, 6f32, 0f32],
+            [0f32, 0f32, 4f32, 5f32],
+        ];
+        let transformed = transform_points(&data, &transform);
+        let expected = VectorData::Vector2(vec![[0f32, 3f32], [4f32, 9f32]]);
+        assert_eq!(expected, transformed)
+    }
+
+    #[test]
+    fn transform_points_vec4() {
+        let data = VectorData::Vector4(vec![[0f32, 1f32, 0f32, -1f32], [2f32, 3f32, 0f32, 5f32]]);
+        let transform = [
+            [2f32, 0f32, 0f32, 0f32],
+            [0f32, 3f32, 0f32, 0f32],
+            [0f32, 0f32, 1f32, 0f32],
+            [0f32, 0f32, 4f32, 5f32],
+        ];
+        let transformed = transform_points(&data, &transform);
+        let expected =
+            VectorData::Vector4(vec![[0f32, 3f32, 4f32, -1f32], [4f32, 9f32, 4f32, 5f32]]);
+        assert_eq!(expected, transformed)
+    }
+
+    #[test]
+    fn transform_vectors_vec2() {
+        let data = VectorData::Vector2(vec![[0f32, 1f32], [2f32, 3f32]]);
+        let transform = [
+            [2f32, 0f32, 0f32, 0f32],
+            [0f32, 3f32, 0f32, 0f32],
+            [0f32, 0f32, 6f32, 0f32],
+            [0f32, 0f32, 4f32, 5f32],
+        ];
+        let transformed = transform_vectors(&data, &transform);
+        let expected = VectorData::Vector2(vec![[0f32, 3f32], [4f32, 9f32]]);
+        assert_eq!(expected, transformed)
+    }
+
+    #[test]
+    fn transform_vectors_vec4() {
+        let data = VectorData::Vector4(vec![[0f32, 1f32, 0f32, -1f32], [2f32, 3f32, 0f32, 5f32]]);
+        let transform = [
+            [2f32, 0f32, 0f32, 0f32],
+            [0f32, 3f32, 0f32, 0f32],
+            [0f32, 0f32, 1f32, 0f32],
+            [0f32, 0f32, 4f32, 5f32],
+        ];
+        let transformed = transform_vectors(&data, &transform);
+        // This is similar to the points test, but the translation should have no effect since w is set to 0.0.
+        let expected =
+            VectorData::Vector4(vec![[0f32, 3f32, 0f32, -1f32], [4f32, 9f32, 0f32, 5f32]]);
+        assert_eq!(expected, transformed)
     }
 }
